# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: http://foglamp.readthedocs.io/
# FOGLAMP_END

""" Face recognition async plugin for Coral"""

import copy
import asyncio
import json
import uuid
import logging
from threading import Thread
import time
import re
import subprocess
import os
import threading
from foglamp.common import logger
from foglamp.plugins.common import utils
import async_ingest

import svgwrite
from foglamp.plugins.south.coral_people import gstreamer
from edgetpu.detection.engine import DetectionEngine

__author__ = "Amandeep Singh Arora"
__copyright__ = "Copyright (c) 2019 Dianomic Systems Inc."
__license__ = "Apache 2.0"
__version__ = "${VERSION}"

_DEFAULT_CONFIG = {
    'plugin': {
        'description': 'Face detection using mobilenet v2 model',
        'type': 'string',
        'default': 'coral_people',
        'readonly': 'true'
    },
    'model_file': {
        'description': 'TFlite model file to use for inference',
        'type': 'string',
        'default': 'mobilenet_ssd_v2_face_quant_postprocess_edgetpu.tflite',
        'order': '1',
        'displayName': 'TFlite model file'
    },
    'labels_file': {
        'description': 'Labels file used during inference',
        'type': 'string',
        'default': 'coco_labels.txt',
        'order': '2',
        'displayName': 'Labels file'
    },
    'asset_name': {
        'description': 'Asset Name',
        'type': 'string',
        'default': 'CoralPeople',
        'order': '3',
        'displayName': 'Asset Name'
    },
    'source': {
        'description': 'Source of video stream (e.g. video0, rtsp://...',
        'type': 'string',
        'default': 'video0',
        'order': '4',
        'displayName': 'Video source'
    }
}


_LOGGER = logger.setup(__name__, level=logging.INFO)


c_callback = None
c_ingest_ref = None
loop = None
t = None
t2 = None
shutdown_in_progress = False
BACKGROUND_TASK = False


def load_labels(path):
    p = re.compile(r'\s*(\d+)(.+)')
    with open(path, 'r', encoding='utf-8') as f:
        lines = (p.match(line).groups() for line in f.readlines())
        return {int(num): text.strip() for num, text in lines}


def shadow_text(dwg, x, y, text, font_size=20):
    dwg.add(dwg.text(text, insert=(x+1, y+1), fill='black', font_size=font_size))
    dwg.add(dwg.text(text, insert=(x, y), fill='white', font_size=font_size))


def generate_svg(dwg, objs, labels, text_lines):
    width, height = dwg.attribs['width'], dwg.attribs['height']
    for y, line in enumerate(text_lines):
        shadow_text(dwg, 10, y*20, line)
    for obj in objs:
        x0, y0, x1, y1 = obj.bounding_box.flatten().tolist()
        x, y, w, h = x0, y0, x1 - x0, y1 - y0
        x, y, w, h = int(x * width), int(y * height), int(w * width), int(h * height)
        percent = int(100 * obj.score)
        label = '%d%% %s' % (percent, labels[obj.label_id])
        shadow_text(dwg, x, y - 5, label)
        dwg.add(dwg.rect(insert=(x, y), size=(w, h), fill='red', fill_opacity=0.3, stroke='white'))


def check_background():
    out = subprocess.Popen(['systemctl' , 'status' , 'foglamp'],
                           stdout=subprocess.PIPE,
                           stderr=subprocess.STDOUT)
    stdout, stderr = out.communicate()
    if str(stdout).find("active (running)") != -1:
        return True
    else:
        return False


def plugin_info():
    """ Returns information about the plugin.
    Args:
    Returns:
        dict: plugin information
    Raises:
    """

    return {
        'name': 'Face recognition plugin',
        'version': '1.8.0',
        'mode': 'async',
        'type': 'south',
        'interface': '1.0',
        'config': _DEFAULT_CONFIG
    }


def plugin_init(config):
    """ Initialise the plugin with the merged config
        Args:
           config: JSON configuration document for the South plugin configuration category
        Returns:
           data: JSON object to be used in future calls to the plugin
        Raises:
    """
    data = copy.deepcopy(config)
    global shutdown_in_progress
    global BACKGROUND_TASK
    BACKGROUND_TASK = check_background()
    shutdown_in_progress = False
    return data


def plugin_start(handle):
    """ Start the plugin operation - observe video feed from camera device and find faces within
        that using the configured ML model
        Args:
            handle: handle returned by the plugin initialisation call
        Returns:
            Sends back the reading object (JSON doc) via C/python async ingest interface when it recognizes a face
        Raises:
    """
    global loop, t, t2
    loop = asyncio.new_event_loop()
    try:
        model = handle['model_file']['value']
        labels = handle['labels_file']['value']
        asset_name = handle['asset_name']['value']
        threshold = 0.1
        top_k = 10

        model = os.path.join(os.path.dirname(__file__), "model", model)
        labels = os.path.join(os.path.dirname(__file__), "model", labels)

        engine = DetectionEngine(model)
        labels = load_labels(labels)

        last_time = time.monotonic()

        def user_callback(image, svg_canvas):

            nonlocal last_time

            start_time = time.monotonic()
            objs = engine.DetectWithImage(image, threshold=threshold,
                                          keep_aspect_ratio=True, relative_coord=True,
                                          top_k=top_k)
            end_time = time.monotonic()
            text_lines = [
                'Inference: %.2f ms' % ((end_time - start_time) * 1000),
                'FPS: %.2f fps' % (1.0/(end_time - last_time)),
            ]
            print(' '.join(text_lines))
            last_time = end_time
            generate_svg(svg_canvas, objs, labels, text_lines)

            coords = []
            width, height = svg_canvas.attribs['width'], svg_canvas.attribs['height']
            for obj in objs:
                x0, y0, x1, y1 = obj.bounding_box.flatten().tolist()
                x0, y0, x1, y1 = int(x0 * width), int(y0 * height), int(x1 * width), int(y1 * height)
                score = int(100 * obj.score)
                coords.append({'corners': [[x0, y0], [x1, y1]], 'score': score})

            reads = {
                'count': len(objs),
                'coords': coords
            }
            data = {
                'asset': asset_name,
                'timestamp': utils.local_timestamp(),
                'key': str(uuid.uuid4()),
                'readings': reads
            }

            if len(objs) > 0:
                _LOGGER.debug(json.dumps(data))
                async_ingest.ingest_callback(c_callback, c_ingest_ref, data)
            pass

        def run():
            global loop
            loop.run_forever()

        def run_gst(user_cb, src):

            while not shutdown_in_progress:
                _ = gstreamer.run_pipeline(user_cb, src, background_task=BACKGROUND_TASK)
                _LOGGER.info('Disconnected from video source, will attempt to reconnect in 5 secs...')

                gstreamer.stop_pipeline()
                time.sleep(5)

        t = Thread(target=run)
        t.start()

        source = handle['source']['value']
        t2 = Thread(target=run_gst, args=(user_callback, source))
        t2.start()
    except Exception as e:
        _LOGGER.exception("Coral People plugin failed to start. Details: %s", str(e))
        raise


def plugin_reconfigure(handle, new_config):
    """ Reconfigures the plugin

    it should be called when the configuration of the plugin is changed during the operation of the south service.
    The new configuration category should be passed.

    Args:
        handle: handle returned by the plugin initialisation call
        new_config: JSON object representing the new configuration category for the category
    Returns:
        new_handle: new handle to be used in the future calls
    Raises:
    """

    # TODO: limit complete plugin restart (shutdown + start) to only the cases where absolutely necessary

    plugin_shutdown(handle)
    new_handle = plugin_init(new_config)

    plugin_start(new_handle)
    return new_handle


def plugin_shutdown(handle):
    """ Shutdowns the plugin doing required cleanup

    To be called prior to the south service being shut down.

    Args:
        handle: handle returned by the plugin initialisation call
    Returns:
    Raises:
    """
    global loop, t, t2, shutdown_in_progress
    try:

        shutdown_in_progress = True
        loop.stop()
        gstreamer.stop_pipeline()

        t2._tstate_lock = None
        t2.stop()
        loop = None
        t = None
        t2 = None
        _LOGGER.info('Coral People plugin shut down.')
    except Exception as e:
        _LOGGER.exception(str(e))
        raise


def plugin_register_ingest(handle, callback, ingest_ref):
    """ Required plugin interface component to communicate to South C server

    Args:
        handle: handle returned by the plugin initialisation call
        callback: C opaque object required to passed back to C/Python async ingest interface
        ingest_ref: C opaque object required to passed back to C/Python async ingest interface
    """
    global c_callback, c_ingest_ref
    c_callback = callback
    c_ingest_ref = ingest_ref
